@@ -1,20 +1,20 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
 
 package IO::Async::SSL;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.14';
+our $VERSION = '0.15';
 $VERSION = eval $VERSION;
 
 use Carp;
 
 use IO::Socket::SSL qw( $SSL_ERROR SSL_WANT_READ SSL_WANT_WRITE );
-use POSIX qw( EAGAIN );
+use POSIX qw( EAGAIN EWOULDBLOCK );
 
 use IO::Async::Handle 0.29;
 use IO::Async::Loop '0.61'; # new Listen API
@@ -60,22 +60,18 @@ the use of SSL or TLS-based connections using L<IO::Socket::SSL>. It does not
 directly provide any methods or functions of its own.
 
 Primarily, it provides C<SSL_connect> and C<SSL_listen>, which yield
-C<IO::Socket::SSL>-upgraded socket handles or L<IO::Async::SSLStream>
+C<IO::Socket::SSL>-upgraded socket handles or L<IO::Async::Stream>
 instances, and two forms of C<SSL_upgrade> to upgrade an existing TCP
 connection to use SSL.
 
 As an additional convenience, if the C<SSL_verify_mode> and C<SSL_ca_*>
-options are omitted, the module will attempt to provide them. If the
-F</etc/ssl/certs> directory exists, it will be used. Failing that, if
-L<Mozilla::CA> can be loaded, that will be used. Otherwise, the module will
-print a warning and set C<SSL_VERIFY_NONE> instead.
+options are omitted, the module will attempt to provide them by quering the
+result of L<IO::Socket::SSL>'s C<default_ca> function. Otherwise, the module
+will print a warning and set C<SSL_VERIFY_NONE> instead.
 
 =cut
 
-# Linux etc.. often stores SSL certs here
-# TODO: Make this a property of IO::Async::OS
-my $SSL_ca_path = "/etc/ssl/certs";
-my %SSL_ca_args;
+my %SSL_ca_args = IO::Socket::SSL::default_ca();
 
 sub _SSL_args
 {
@@ -84,21 +80,9 @@ sub _SSL_args
    # SSL clients (i.e. non-server) require a verify mode
    if( !$args{SSL_server} and !defined $args{SSL_verify_mode} and
        !defined $args{SSL_ca_file} and !defined $args{SSL_ca_path} ) {
-      # Try to load Mozilla::CA; but if it fails remember that so we don't
-      # reload it repeatedly
       unless( %SSL_ca_args ) {
-         if( -d $SSL_ca_path ) {
-            $SSL_ca_args{SSL_verify_mode} = IO::Socket::SSL::SSL_VERIFY_PEER();
-            $SSL_ca_args{SSL_ca_path}     = $SSL_ca_path;
-         }
-         elsif( eval { require Mozilla::CA } ) {
-            $SSL_ca_args{SSL_verify_mode} = IO::Socket::SSL::SSL_VERIFY_PEER();
-            $SSL_ca_args{SSL_ca_file}     = Mozilla::CA::SSL_ca_file();
-         }
-         else {
-            carp "Unable to set SSL_VERIFY_PEER because Mozilla::CA is unavailable";
-            $SSL_ca_args{SSL_verify_mode} = IO::Socket::SSL::SSL_VERIFY_NONE();
-         }
+         carp "Unable to set SSL_VERIFY_PEER because IO::Socket::SSL::default_ca() gives nothing";
+         $SSL_ca_args{SSL_verify_mode} = IO::Socket::SSL::SSL_VERIFY_NONE();
       }
 
       %args = ( %SSL_ca_args, %args );
@@ -107,6 +91,44 @@ sub _SSL_args
    return %args;
 }
 
+sub sslread
+{
+   my $stream = shift;
+   my ( $fh, undef, $len ) = @_;
+
+   my $ret = $stream->_sysread( $fh, $_[1], $len );
+
+   my $read_wants_write = !defined $ret &&
+                          ( $! == EAGAIN or $! == EWOULDBLOCK ) &&
+                          $SSL_ERROR == SSL_WANT_WRITE;
+   $stream->want_writeready_for_read( $read_wants_write );
+
+   # It's possible SSL_read took all the data out of the filehandle, thus
+   # making it not appear read-ready any more.
+   if( $fh->pending ) {
+      $stream->loop->later( sub { $stream->on_read_ready } );
+   }
+
+   return $ret;
+}
+
+sub sslwrite
+{
+   my $stream = shift;
+   my ( $fh, $buf, $len ) = @_;
+
+   my $ret = $stream->_syswrite( $fh, $_[1], $len );
+
+   my $write_wants_read = !defined $ret &&
+                          ( $! == EAGAIN or $! == EWOULDBLOCK ) &&
+                          $SSL_ERROR == SSL_WANT_READ;
+   $stream->want_readready_for_write( $write_wants_read );
+   # If write wants read, there's no point waiting on writereadiness either
+   $stream->want_writeready_for_write( !$write_wants_read );
+
+   return $ret;
+}
+
 =head1 LOOP METHODS
 
 The following extra methods are added to L<IO::Async::Loop>.
@@ -152,8 +174,8 @@ returned future is not used:
 =item on_upgraded => CODE
 
 A continuation that is invoked when the socket has been successfully upgraded
-to SSL. It will be passed an instance of an C<IO::Socket::SSL>, which must be
-wrapped in an instance of L<IO::Async::SSLStream>.
+to SSL. It will be passed an instance of an C<IO::Socket::SSL>, which will
+have appropriate SSL-compatible reader/writer functions attached.
 
  $on_upgraded->( $sslsocket )
 
@@ -218,8 +240,8 @@ sub IO::Async::Loop::SSL_upgrade
          if( $stream ) {
             $stream->configure(
                handle => $socket,
-               reader => \&IO::Async::SSLStream::sslread,
-               writer => \&IO::Async::SSLStream::sslwrite,
+               reader => \&sslread,
+               writer => \&sslwrite,
             );
          }
 
@@ -227,7 +249,7 @@ sub IO::Async::Loop::SSL_upgrade
          return;
       }
 
-      if( $! != EAGAIN ) {
+      if( $! != EAGAIN and $! != EWOULDBLOCK ) {
          my $errstr = IO::Socket::SSL::errstr();
          $loop->remove( $self );
          $f->fail( $errstr, "ssl" );
@@ -321,9 +343,10 @@ sub IO::Async::Loop::SSL_connect
    my $on_ssl_error = delete $params{on_ssl_error} or defined wantarray or
       croak "Expected 'on_ssl_error' or to return a Future";
 
-   require IO::Async::SSLStream;
-
-   my $stream = delete $params{handle} || IO::Async::Stream->new;
+   my $stream = delete $params{handle} || do {
+      require IO::Async::Stream;
+      IO::Async::Stream->new;
+   };
 
    $stream->isa( "IO::Async::Stream" ) or
       croak "Can only SSL_connect a handle instance of IO::Async::Stream";
@@ -417,16 +440,18 @@ sub IO::Async::Loop::SSL_listen
          $listener->$cleartext_acceptor( $listen_sock )->then( sub {
             my ( $socket ) = @_;
 
+            return Future->done() unless $socket; # EAGAIN
+
             $stream->configure( handle => $socket ) if $stream;
 
             $loop->SSL_upgrade(
                _SSL_args( SSL_server => 1, %ssl_params ),
                handle   => ( $stream || $socket ),
-            )
-         })->else( sub {
-            my ( $failure ) = @_;
-            $on_ssl_error->( $failure ) if $on_ssl_error;
-            return Future->new->fail( $failure, ssl => );
+            )->else( sub {
+               my ( $failure ) = @_;
+               $on_ssl_error->( $failure ) if $on_ssl_error;
+               return Future->new->fail( $failure, ssl => );
+            });
          });
       };
 
@@ -452,10 +477,10 @@ The following extra methods are added to L<IO::Async::Protocol::Stream>.
 =head2 $protocol->SSL_upgrade( %params )
 
 A shortcut to calling C<< $loop->SSL_upgrade >>. This method will unconfigure
-the C<transport> of the Protocol, upgrade it to SSL, then reconfigure it again
-newly wrapped in a C<IO::Async::SSLStream> instance. It takes the same
-arguments as C<< $loop->SSL_upgrade >>, except that the C<handle> argument is
-not required as it's taken from the Protocol's C<transport>.
+the C<transport> of the Protocol, upgrade its underlying filehandle to SSL,
+then reconfigure it again with SSL reader and writer functions on it. It takes
+the same arguments as C<< $loop->SSL_upgrade >>, except that the C<handle>
+argument is not required as it's taken from the Protocol's C<transport>.
 
 =cut
 
@@ -468,22 +493,16 @@ sub IO::Async::Protocol::Stream::SSL_upgrade
 
    my $loop = $protocol->get_loop or croak "Expected to be a member of a Loop";
 
-   require IO::Async::SSLStream;
-
-   my $socket = $protocol->transport->read_handle;
+   my $transport = $protocol->transport;
 
    $protocol->configure( transport => undef );
 
    $loop->SSL_upgrade(
-      handle => $socket,
+      handle => $transport,
       on_upgraded => sub {
-         my ( $newsocket ) = @_;
+         my ( $transport ) = @_;
 
-         my $sslstream = IO::Async::SSLStream->new(
-            handle => $newsocket,
-         );
-
-         $protocol->configure( transport => $sslstream );
+         $protocol->configure( transport => $transport );
 
          $on_upgraded->();
       },
