@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2013 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
 
 package IO::Async::SSLStream;
 
@@ -10,11 +10,10 @@ use warnings;
 use base qw( IO::Async::Stream );
 IO::Async::Stream->VERSION( '0.59' );
 
-our $VERSION = '0.14';
+our $VERSION = '0.15';
 $VERSION = eval $VERSION;
 
-use IO::Socket::SSL qw( $SSL_ERROR SSL_WANT_READ SSL_WANT_WRITE );
-use POSIX qw( EAGAIN );
+use IO::Async::SSL;
 
 =head1 NAME
 
@@ -40,46 +39,12 @@ sub _init
    my $self = shift;
    my ( $params ) = @_;
 
-   $params->{reader} = \&sslread;
-   $params->{writer} = \&sslwrite;
+   $params->{reader} = \&IO::Async::SSL::sslread;
+   $params->{writer} = \&IO::Async::SSL::sslwrite;
 
    $self->SUPER::_init( $params );
 }
 
-sub sslread
-{
-   my $stream = shift;
-   my ( $fh, undef, $len ) = @_;
-
-   my $ret = $stream->_sysread( $fh, $_[1], $len );
-
-   my $read_wants_write = !defined $ret && $! == EAGAIN && $SSL_ERROR == SSL_WANT_WRITE;
-   $stream->want_writeready_for_read( $read_wants_write );
-
-   # It's possible SSL_read took all the data out of the filehandle, thus
-   # making it not appear read-ready any more.
-   if( $fh->pending ) {
-      $stream->loop->later( sub { $stream->on_read_ready } );
-   }
-
-   return $ret;
-}
-
-sub sslwrite
-{
-   my $stream = shift;
-   my ( $fh, $buf, $len ) = @_;
-
-   my $ret = $stream->_syswrite( $fh, $_[1], $len );
-
-   my $write_wants_read = !defined $ret && $! == EAGAIN && $SSL_ERROR == SSL_WANT_READ;
-   $stream->want_readready_for_write( $write_wants_read );
-   # If write wants read, there's no point waiting on writereadiness either
-   $stream->want_writeready_for_write( !$write_wants_read );
-
-   return $ret;
-}
-
 =head1 BUGS
 
 =over 4
